import asyncio
import logging
from random import randint

import websockets
from tapioca_gather import Gather

from .actions import set_actions_methods
from .events_pb2 import ServerClientBatch
from .helpers import get_logger


def get_txn_id() -> int:
    return randint(0, 4294967295)


class GatherClient:
    def __init__(
        self,
        api_key,
        space_id,
        callback=None,
        log_level=logging.INFO,
        hearbeat_interval=15,
    ):
        self._api_key = api_key
        self._space_id = space_id.replace("/", "\\")
        self._callback = callback
        self._ws = None
        self._logger = get_logger(log_level=log_level)
        self._heartbeat_interval = hearbeat_interval
        self._user_id = None
        set_actions_methods(self)

    async def _consumer(self):
        if self._ws is None or not self._ws.open:
            self._logger.warning("Not connected")
            return
        self._logger.debug("Receiving messages")
        while True:
            message = await self._ws.recv()
            server_response = ServerClientBatch()
            server_response.ParseFromString(message)
            if self._callback is not None:
                await self._callback(self, server_response)
            self._logger.debug(f"Message received: {server_response}")

    async def _heartbeat(self):
        if self._ws is None or not self._ws.open:
            self._logger.warning("Not connected")
            return
        while True:
            await self.client_heartbeat()
            self._logger.debug("Heartbeat sent")
            await asyncio.sleep(self._heartbeat_interval)

    async def _connect(self):
        api = Gather(api_key=self._api_key)
        url = api.game_server_assignment(space_id=self._space_id).get()().data
        while True:
            try:
                self._ws = await websockets.connect(url)
                self._logger.info(f"Connected to {url}")
                break
            except Exception as e:
                self._logger.error(f"Conenction error: {e}")
                await asyncio.sleep(3)
                self._logger.debug("Retrying to connect")
        await self.init(space_id=self._space_id, api_key=self._api_key)
        message = await self._ws.recv()
        server_response = ServerClientBatch()
        server_response.ParseFromString(message)
        if len(server_response.events) == 0:
            self._logger.error("Connection not ready")
            await self._ws.close()
            return
        event = server_response.events[0]
        if event.WhichOneof("event") != "ready":
            self._logger.error("Connection not ready")
            await self._ws.close()
            return
        self._user_id = event.ready.id
        self._logger.info(f"Connected as user {self._user_id}")

    async def run(self, producer, *args, **kwargs):
        await self._connect()
        _, pending = await asyncio.wait(
            [
                asyncio.create_task(self._consumer()),
                asyncio.create_task(self._heartbeat()),
                asyncio.create_task(producer(self, *args, **kwargs)),
            ],
            return_when=asyncio.FIRST_COMPLETED,
        )
        if self._ws is not None:
            await self._ws.close()
        for task in pending:
            task.cancel()



import logging
import re

_CAMEL_PATTERN = re.compile(r"(?<!^)(?=[A-Z])")


def get_logger(name="gather_client_ws", log_level=logging.INFO):
    logger = logging.getLogger(name)
    logger.setLevel(log_level)
    formatter = logging.Formatter("[%(asctime)s][%(levelname)s] %(message)s")
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)
    return logger


def camel_to_snake(s):
    return _CAMEL_PATTERN.sub("_", s).lower()


def snake_to_camel(s):
    pascal_case = "".join(word.title() for word in s.split("_"))
    return pascal_case[0].lower() + pascal_case[1:]



import types
from random import randint

from gather_client_ws.helpers import camel_to_snake, snake_to_camel

from .events_pb2 import ClientServerAction


def get_txn_id():
    return randint(0, 4294967295)


def get_type(descriptor):
    if descriptor.type == descriptor.TYPE_BOOL:
        return bool
    if descriptor.type == descriptor.TYPE_STRING:
        return str
    if descriptor.type == descriptor.TYPE_BYTES:
        return bytes
    if descriptor.type in [descriptor.TYPE_DOUBLE, descriptor.TYPE_FLOAT]:
        return float
    return int


def get_available_actions():
    actions = []
    fields = ClientServerAction.DESCRIPTOR.fields_by_name
    for field_name, field_descriptor in fields.items():
        # The only primitive field is the txnId, it is safe to ignore it here
        if field_descriptor.message_type is None:
            continue
        action_args = []
        action_descriptor = field_descriptor.message_type
        for name, descriptor in action_descriptor.fields_by_name.items():
            if descriptor.message_type is None:  # is primitive
                action_args.append({name: get_type(descriptor)})
                continue
            action_args.append({name: descriptor.message_type._concrete_class})
        actions.append(
            {
                "name": field_name,
                "args": action_args,
                "class": field_descriptor.message_type._concrete_class,
            }
        )
    return actions


def create_action_method(action):
    async def method(client, *args, **kwargs):
        params = {}
        for value, arg in zip(args, action["args"]):
            params[next(iter(arg))] = value
        for key, value in kwargs.items():
            params[snake_to_camel(key)] = value
        data = action["class"](**params)
        _action = ClientServerAction(
            txnId=get_txn_id(), **{action["name"]: data}
        )
        await client._ws.send(_action.SerializeToString())

    args_doc = "\n".join(
        f"\t{camel_to_snake(name)} ({_type.__name__})"
        for arg in action["args"]
        for name, _type in arg.items()
    )

    method.__doc__ = f"{camel_to_snake(action['name'])}\nArgs:\n{args_doc}"

    return method


def set_actions_methods(client):
    actions = get_available_actions()
    actions_names = []
    for action in actions:
        name = camel_to_snake(action["name"])
        actions_names.append(name)
        setattr(
            client,
            name,
            types.MethodType(create_action_method(action), client),
        )
    client.available_actions = actions_names
